from unittest.mock import patch, MagicMock
from django.test import override_settings
from rest_framework.reverse import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.models import DataImport
from arkindex.documents.models import Element, ElementType, Corpus
from arkindex.project.tests import FixtureAPITestCase


class TestWorkflows(FixtureAPITestCase):
    """
    Test workflows from Arkindex existing elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.volume = Element.objects.get(type=ElementType.Volume, name='Volume 1')
        cls.pages = Element.objects.get_descending(cls.volume.id).filter(type=ElementType.Page)
        cls.private_volume = cls.private_corpus.elements.create(type=ElementType.Volume, name='Hidden')

    @override_settings(
        ARKINDEX_TASKS_IMAGE='tasks:latest',
        PONOS_RECIPE={'env': {'SOME_VAR': 'somevalue'}}
    )
    @patch('arkindex_common.ml_tool.MLTool.get')
    def test_volume_workflow(self, ml_get_mock):
        dla_mock, reco_mock = MagicMock(), MagicMock()
        dla_mock.type, dla_mock.slug = MLToolType.DLAnalyser, 'line-detector'
        reco_mock.type, reco_mock.slug = MLToolType.Recognizer, 'text-recognizer'
        dla_mock.task_recipe = {'image': 'im1', 'command': 'cmd1'}
        reco_mock.task_recipe = {'image': 'im2', 'command': 'cmd2'}
        tools = {dla_mock.slug: dla_mock, reco_mock.slug: reco_mock}
        ml_get_mock.side_effect = lambda directory, type, slug: tools[slug]

        self.client.force_login(self.user)
        ml_tools = [
            {'type': 'dla', 'slug': 'line-detector', 'parents': []},
            {
                'type': 'recognizer',
                'slug': 'text-recognizer',
                'parents': [{'type': 'dla', 'slug': 'line-detector'}],
            }
        ]

        response = self.client.post(
            reverse('api:element-workflow', kwargs={'pk': self.volume.id}),
            ml_tools,
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertDictEqual(data, {
            'corpus': str(self.volume.corpus.id),
            'files': [],
            'id': str(dataimport.id),
            'mode': 'elements',
            'payload': {
                'element_id': str(self.volume.id),
                'ml_tools': [
                    {
                        'slug': 'line-detector',
                        'type': 'dla',
                        'parents': [],
                    }, {
                        'slug': 'text-recognizer',
                        'type': 'recognizer',
                        'parents': [{'slug': 'line-detector', 'type': 'dla'}]
                    }
                ],
                'volume_id': str(self.volume.id)
            },
            'revision': dataimport.revision,
            'state': 'unscheduled',
            'workflow': reverse(
                'ponos:workflow-details',
                kwargs={'pk': workflow.id},
                request=response.wsgi_request
            )
        })
        self.assertEqual(workflow.recipes['import'].image, 'tasks:latest')
        self.assertEqual(
            workflow.recipes['import'].command,
            'python -m arkindex_tasks.init_elements {}'.format(str(self.volume.id))
        )

    def test_page_workflow(self):
        page = self.pages[0]
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-workflow', kwargs={'pk': page.id}),
            [],
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertDictEqual(data, {
            'corpus': str(page.corpus.id),
            'files': [],
            'id': str(dataimport.id),
            'mode': 'elements',
            'payload': {
                'element_id': str(page.id),
                'ml_tools': [],
                'volume_id': str(self.volume.id)
            },
            'revision': dataimport.revision,
            'state': 'unscheduled',
            'workflow': reverse(
                'ponos:workflow-details',
                kwargs={'pk': workflow.id},
                request=response.wsgi_request
            )
        })

    def test_element_workflow_requires_login(self):
        ml_tools = [{'type': 'dla', 'slug': 'line-detector', 'parents': []}]
        response = self.client.post(
            reverse('api:element-workflow', kwargs={'pk': self.private_volume.id}),
            ml_tools,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_workflow_acl(self):
        self.client.force_login(self.user)
        ml_tools = [{'type': 'dla', 'slug': 'line-detector', 'parents': []}]
        response = self.client.post(
            reverse('api:element-workflow', kwargs={'pk': self.private_volume.id}),
            ml_tools,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_wrong_element_type(self):
        self.client.force_login(self.user)
        ml_tools = [{'type': 'dla', 'slug': 'line-detector', 'parents': []}]
        act = Element.objects.filter(type=ElementType.Act).first()
        response = self.client.post(
            reverse('api:element-workflow', kwargs={'pk': act.id}),
            ml_tools,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
