from django.db.models import Prefetch, prefetch_related_objects
from django_filters import rest_framework as filters
from rest_framework.exceptions import ValidationError
from rest_framework.generics import (
    CreateAPIView, ListAPIView, ListCreateAPIView,
    RetrieveAPIView, RetrieveUpdateAPIView, RetrieveUpdateDestroyAPIView
)

from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import \
    Act, Classification, Corpus, Element, ElementType, Page, Right, Transcription, Region
from arkindex.documents.serializers.elements import (
    ActSerializer, CorpusSerializer, ElementCreateSerializer, ElementSerializer, ElementSlimSerializer,
    PageSerializer, SurfaceSerializer, RegionSerializer, RegionCreateSerializer, ElementNeighborsSerializer
)
from arkindex.documents.serializers.light import PageLightSerializer, RegionLightSerializer
from arkindex.documents.serializers.ml import TranscriptionSerializer
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.pagination import PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly

from .filters import ElementsMLClassFilter


classifications_prefetch = Prefetch(
    'classifications',
    queryset=Classification.objects.select_related('ml_class', 'source').order_by('-confidence')
)


class ElementsList(CorpusACLMixin, ListAPIView):
    """
    List documents using pagination
    """
    serializer_class = ElementSlimSerializer
    queryset = Element.objects.all()
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = ElementsMLClassFilter

    def filter_queryset(self, queryset):
        filters = {
            'type': self.request.query_params.get('type', ElementType.Volume),
        }
        if not isinstance(filters['type'], ElementType):
            try:
                filters['type'] = ElementType(filters['type'])
            except ValueError as e:
                raise ValidationError({'type': [str(e)]})

        if 'corpus' in self.request.query_params:
            filters['corpus'] = self.get_corpus(self.request.query_params['corpus'])
        else:
            filters['corpus__in'] = Corpus.objects.readable(self.request.user)

        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.request.query_params['name']

        return super().filter_queryset(queryset) \
            .filter(**filters) \
            .prefetch_related('corpus') \
            .order_by('corpus', 'type', 'name')


class ElementRetrieve(RetrieveUpdateAPIView):
    """
    Retrieve a single element
    """
    serializer_class = ElementSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        if self.request.method == 'GET':
            corpora = Corpus.objects.readable(self.request.user)
        else:
            corpora = Corpus.objects.writable(self.request.user)
        return Element.objects \
            .filter(corpus__in=corpora) \
            .select_related('corpus') \
            .prefetch_related('metadatas__revision__repo', 'metadatas__entity', classifications_prefetch)


class RelatedElementsList(ListAPIView):
    """
    List all documents linked to a specified document
    """
    serializer_class = ElementSlimSerializer
    # Tell the OpenAPI schema generator to believe this view is a list
    action = 'List'

    def get_queryset(self):
        filtering = {
            'corpus__in': Corpus.objects.readable(self.request.user),
        }
        type_param = self.request.query_params.get('type')
        if type_param is not None:
            filtering['type'] = type_param
        related = Element.objects.get_related(self.kwargs['pk'], **filtering)
        prefetch_related_objects(related, 'corpus')
        return related


class ElementNeighbors(ListAPIView):
    """
    List neighboring elements
    """
    serializer_class = ElementNeighborsSerializer

    def get_queryset(self):

        def presentation(n):
            parents = Element.objects.filter(id__in=n.path).order_by('name')
            return {'position': n.ordering, 'element': n.element, 'parents': parents}

        n = int(self.request.query_params.get('n', 1))
        element = Element.objects.get(id=self.kwargs['pk'])
        neighbors = Element.objects.get_neighbors(element, n)
        return [presentation(n) for n in neighbors]


class ElementPages(ListAPIView):
    """
    List pages inside an element
    """
    serializer_class = PageLightSerializer

    def get_queryset(self):
        return Page.objects.get_descending(
            self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user),
            zone__isnull=False,
        ).prefetch_related('zone__image__server', classifications_prefetch)


class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)


class CorpusRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Get, update and delete corpora
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if request.method != 'GET' and Right.Write not in obj.get_acl_rights(request.user):
            self.permission_denied(request, message='You do not have write access to this corpus.')


class PageDetails(RetrieveAPIView):
    """
    Get details for a specific page
    """
    serializer_class = PageSerializer

    def get_queryset(self):
        return Page.objects \
                   .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
                   .select_related('zone__image__server', 'corpus') \
                   .prefetch_related('classifications__source', 'transcriptions__source')

    def get_object(self):
        page = super().get_object()
        page.parent_elements = Element.objects.get_ascending_paths(page.id)
        return page


class SurfaceDetails(RetrieveAPIView):
    """
    Get details for a specific surface
    """
    serializer_class = SurfaceSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=ElementType.Surface,
            corpus__in=Corpus.objects.readable(self.request.user),
        ).select_related('zone__image__server')

    def get_object(self):
        surf = super().get_object()
        surf.page = Page.objects.filter(zone__image=surf.zone.image) \
            .select_related('zone__image__server').first()
        return surf


# TODO: Edition disabled
class ActEdit(RetrieveAPIView):
    """
    Edit acts and associated ElementPaths
    """
    serializer_class = ActSerializer

    def get_queryset(self):
        return Act.objects \
            .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
            .prefetch_related('metadatas__revision', 'metadatas__entity', classifications_prefetch)

    def get_object(self):
        act = super().get_object()
        act.parent_elements = Element.objects.get_ascending_paths(act.id)
        act.child_elements = Element.objects.get_descending(act.id)
        return act


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


class ElementTranscriptions(ListAPIView):
    """
    Get transcriptions corresponding to a specific element
    Transcriptions can be filtered by type
    """
    serializer_class = TranscriptionSerializer
    pagination_class = TranscriptionsPagination

    def get_queryset(self):
        queryset = Transcription.objects.filter(
            element_id=self.kwargs['pk'],
            element__corpus__in=Corpus.objects.readable(self.request.user),
        ).prefetch_related('zone__image__server', 'source').order_by('id')
        req_type = self.request.query_params.get('type')
        if req_type:
            try:
                req_type = TranscriptionType(req_type)
                queryset = queryset.filter(type=req_type)
            except ValueError:
                raise ValidationError({'type': 'Not a valid transcription type'})
        return queryset


class ElementRegions(ListAPIView):
    serializer_class = RegionLightSerializer

    def get_queryset(self):
        return Region.objects \
            .filter(
                elements=self.kwargs['pk'],
                elements__corpus__in=Corpus.objects.readable(self.request.user),
            ) \
            .select_related('ml_class', 'source') \
            .prefetch_related('zone__image__server') \
            .order_by('region_elements__ordering')


class ElementsCreate(CreateAPIView):
    """
    Add a child to an existing element
    Create a new Element
    A parent and/or an image can be attached to this Element
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementCreateSerializer


class RegionDetails(RetrieveAPIView):
    serializer_class = RegionSerializer

    def get_queryset(self):
        return Region.objects \
            .filter(elements__corpus__in=Corpus.objects.readable(self.request.user)) \
            .select_related('ml_class', 'source') \
            .prefetch_related('zone__image__server', 'elements')


class RegionCreate(CreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = RegionCreateSerializer

    def get_queryset(self):
        return Element.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:  # Ignore this step when generating the schema with OpenAPI
            context['element'] = self.get_object()
        return context
