from django.conf import settings
from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.mixins import CreateModelMixin
from rest_framework.generics import GenericAPIView, CreateAPIView, UpdateAPIView
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView
from arkindex.documents.models import \
    Classification, ClassificationState, DataSource, Transcription, Page, Corpus, Entity
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import TranscriptionType, EntityType
from arkindex.documents.serializers.ml import (
    ClassificationsSerializer, ClassificationCreateSerializer, ClassificationSerializer,
    TranscriptionsSerializer, TranscriptionCreateSerializer
)
from arkindex.documents.indexer import Indexer
from arkindex.documents.pagexml import PageXmlParser
from arkindex.images.models import Zone
from arkindex.images.importer import build_transcriptions, save_transcriptions, index_transcriptions
from arkindex.project.parsers import XMLParser
from arkindex.project.permissions import IsVerified
from arkindex.project.polygon import Polygon
import os.path
import logging

logger = logging.getLogger(__name__)


class TranscriptionCreate(CreateAPIView):
    serializer_class = TranscriptionCreateSerializer
    permission_classes = (IsVerified, )

    def perform_create(self, serializer):
        transcription_type = serializer.validated_data['type']
        if transcription_type not in TranscriptionType:
            raise ValidationError("This endpoint can only import transcriptions.")
        element = serializer.validated_data['element']

        # Create a zone on the page's image
        polygon = Polygon(serializer.validated_data['polygon'])
        try:
            ts_zone, _ = Zone.objects.get_or_create(
                image=element.zone.image,
                polygon=polygon,
            )
        except Zone.MultipleObjectsReturned:
            # This should not happend, but we cannot set a unique
            # on polygon PG fields
            ts_zone = Zone.objects.filter(
                image=element.zone.image,
                polygon=polygon,
            ).first()

        ts, created = Transcription.objects.get_or_create(
            element=element,
            source=serializer.validated_data['source'],
            type=transcription_type,
            zone=ts_zone,
            text=serializer.validated_data['text'],
            defaults={
                'score': serializer.validated_data['score']
            }
        )
        if not created or ts.score != serializer.validated_data['score']:
            ts.score = serializer.validated_data['score']
            ts.save()

        # Index in ES
        Indexer().run_index(
            Transcription.objects.filter(id=ts.id)
        )
        return ts

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response({'id': obj.id}, status=status.HTTP_201_CREATED, headers=headers)


class TranscriptionBulk(CreateAPIView, UpdateAPIView):
    '''
    Create transcriptions in bulk, all linked to the same image
    and parent element
    '''
    serializer_class = TranscriptionsSerializer
    permission_classes = (IsVerified, )
    # Force DRF to ignore PATCH
    http_method_names = ['post', 'put', 'head', 'options', 'trace']

    def get_object(self):  # Ignore the get_object part in UpdateAPIView.update
        return

    def perform_create(self, serializer):
        self.run(serializer, delete=False)

    def perform_update(self, serializer):
        self.run(serializer, delete=True)

    def run(self, serializer, delete=False):
        parent = serializer.validated_data['parent']
        source = DataSource.from_ml_tool(serializer.validated_data['recognizer'])
        if delete:
            # Clear all previous transcriptions from source
            parent.transcriptions.filter(source=source).delete()

        trpolygons = build_transcriptions(
            image=parent.zone.image,
            parent=parent,
            source=source,
            items=[
                {
                    'polygon': tr['polygon'],
                    'text': tr['text'],
                    'score': tr['score'],
                    'type': tr['type'],
                }
                for tr in serializer.validated_data['transcriptions']
            ],
        )
        if not trpolygons:
            return

        transcriptions, _ = save_transcriptions(*trpolygons)

        # Build the Page transcription
        text = '\n'.join(
            t.text
            for t in sorted(
                parent.transcriptions.filter(
                    type=TranscriptionType.Line,
                    source=source,
                ).prefetch_related('zone'),
                key=lambda t: (t.zone.polygon.center.y, t.zone.polygon.center.x),
            )
        )

        # Add the (page transcription ID, element ID) tuple to the transcriptions tuple
        transcriptions += (
            (
                parent.transcriptions.create(
                    type=TranscriptionType.Page,
                    text=text,
                    zone=parent.zone,
                    source=source,
                ).id,
                parent.id,
            ),
        )
        index_transcriptions(transcriptions)


class ClassificationCreate(CreateAPIView):
    """
    Create a classification for a specific element
    """
    serializer_class = ClassificationCreateSerializer
    queryset = Classification.objects.all()

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data_source, __ = DataSource.objects.get_or_create(
            type=MLToolType.Classifier,
            slug='manual',
            defaults={
                'revision': '',
                'internal': False,
            }
        )
        serializer.save(
            source=data_source,
            moderator=self.request.user,
            state=ClassificationState.Validated,
            confidence=1
        )
        headers = self.get_success_headers(serializer.data)
        response_serializer = ClassificationSerializer(instance=serializer.instance)
        return Response(response_serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class ClassificationBulk(CreateAPIView):
    """
    Create multiple classifications at once for a single element from a single ML tool
    """
    serializer_class = ClassificationsSerializer
    permission_classes = (IsVerified, )

    def perform_create(self, serializer):
        parent = serializer.validated_data['parent']
        source = DataSource.from_ml_tool(serializer.validated_data['classifier'])
        parent.classifications.filter(source=source).delete()
        Classification.objects.bulk_create([
            Classification(
                element=parent,
                source=source,
                ml_class=cl['ml_class'],
                confidence=cl['confidence'],
            )
            for cl in serializer.validated_data['classifications']
        ])


class ClassificationModerationActionsMixin(GenericAPIView):
    serializer_class = ClassificationSerializer

    def get_queryset(self):
        return Classification.objects.filter(element__corpus__in=Corpus.objects.writable(self.request.user))

    def put(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.moderator = self.request.user
        instance.state = self.moderation_action
        instance.save(update_fields=['moderator', 'state'])
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_202_ACCEPTED)


class ClassificationValidate(ClassificationModerationActionsMixin):
    """Reject a classification"""
    moderation_action = ClassificationState.Validated
    action = 'Validate'  # For OpenAPI operationId -> ValidateClassification


class ClassificationReject(ClassificationModerationActionsMixin):
    """Reject a classification"""
    moderation_action = ClassificationState.Rejected
    action = 'Reject'  # For OpenAPI operationId -> RejectClassification


class PageXmlTranscriptionsImport(CreateModelMixin, APIView):
    parser_classes = (XMLParser, )
    permission_classes = (IsVerified, )

    def get_queryset(self):
        return Page.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_object(self):
        """
        Since we are inheriting from APIView, because GenericAPIView would break OpenAPI,
        we have to rewrite get_object ourselves.
        """
        obj = get_object_or_404(self.get_queryset(), pk=self.kwargs['pk'])
        self.check_object_permissions(self.request, obj)
        return obj

    def post(self, request, *args, **kwargs):
        page = self.get_object()
        try:
            parser = PageXmlParser(request.data)
        except AssertionError as e:
            raise ValidationError('Could not parse PAGE XML document: {!s}'.format(e))

        transcriptions = parser.save_transcriptions(page)

        blocks = parser.create_blocks(transcriptions)
        blocks = parser.merge(blocks)

        path = os.path.join(settings.BASE_DIR, 'documents', 'fixtures', 'roles_transkribus.yml')
        types = {
            '_nom': EntityType.Person.value,
            '_prenom': EntityType.Person.value,
            '_lieu_residence': EntityType.Location.value,
            '_profession': EntityType.Subject.value,
        }
        entities_id = parser.create_objects(blocks, page, path, types)
        logger.info('{} blocks: {} entities created'.format(
            len(blocks),
            len(entities_id)
        ))

        if entities_id:
            # Reindex entities to add new entities in ElasticSearch
            Indexer().run_index(
                Entity.objects.filter(id__in=entities_id)
            )

        return Response(
            status=status.HTTP_201_CREATED,
            headers=self.get_success_headers(None),
        )
