#!/usr/bin/env python3
from django.core.management.base import BaseCommand
from django.contrib.auth.models import Group
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import TranscriptionType, MetaType
from arkindex.documents.models import Corpus, Element, ElementType, Transcription, DataSource, MetaData
from arkindex.images.models import ImageServer, Image, Zone
from arkindex.users.models import User, CorpusRight


def makezone(img, x1, x2):
    """
    Shortcut to create a square zone on an image
    """
    return Zone.objects.create(polygon=[(x1, x1), (x2, x1), (x2, x2), (x1, x2), (x1, x1)], image=img)


class Command(BaseCommand):
    help = 'Create test objects in the database'

    def handle(self, *args, **options):
        # Create 1 image server
        imgsrv = ImageServer.objects.create(
            display_name="Test Server",
            url="http://server",
            s3_bucket='iiif',
        )

        # Create 6 images
        img1 = Image.objects.create(path='img1', width=1000, height=1000, server=imgsrv)
        img2 = Image.objects.create(path='img2', width=1000, height=1000, server=imgsrv)
        img3 = Image.objects.create(path='img3', width=1000, height=1000, server=imgsrv)
        img4 = Image.objects.create(path='img4', width=1000, height=1000, server=imgsrv)
        img5 = Image.objects.create(path='img5', width=1000, height=1000, server=imgsrv)
        img6 = Image.objects.create(path='img6', width=1000, height=1000, server=imgsrv)

        # Create 6 zones on the images for pages
        z1 = makezone(img1, 0, 1000)
        z2 = makezone(img2, 0, 1000)
        z3 = makezone(img3, 0, 1000)
        z4 = makezone(img4, 0, 1000)
        z5 = makezone(img5, 0, 1000)
        z6 = makezone(img6, 0, 1000)

        # Create internal and demo user groups
        # They may have already been created by users.0002_default_groups
        Group.objects.get_or_create(id=1, name='Demo')
        Group.objects.get_or_create(id=2, name='Internal')

        # Create an admin, a user and a demo user
        superuser = User.objects.create_superuser('root@root.fr', 'Pa$$w0rd')
        superuser.verified_email = True
        superuser.save()

        user = User.objects.create_user('user@user.fr', 'Pa$$w0rd')
        user.verified_email = True
        user.save()

        User.objects.create_demo_user('demo@demo.fr')

        # Create 1 data source for transcriptions
        source = DataSource.objects.create(
            type=MLToolType.Recognizer,
            slug='test',
            revision='4.2',
            internal=False,
        )

        # Create 1 corpus
        corpus = Corpus.objects.create(name='Unit Tests', public=True)

        # Give rights to the user
        CorpusRight.objects.create(user=user, corpus=corpus, can_write=True, can_admin=True)

        # Create 2 volumes
        vol1 = Element.objects.create(corpus=corpus, type=ElementType.Volume, name="Volume 1")
        vol2 = Element.objects.create(corpus=corpus, type=ElementType.Volume, name="Volume 2")

        # Create 3 pages in each volume
        p1_1 = Element.objects.create(corpus=corpus, type=ElementType.Page, name="Volume 1, page 1r", zone=z1)
        p1_2 = Element.objects.create(corpus=corpus, type=ElementType.Page, name="Volume 1, page 1v", zone=z2)
        p1_3 = Element.objects.create(corpus=corpus, type=ElementType.Page, name="Volume 1, page 2r", zone=z3)
        p1_1.metadatas.set([MetaData(name='folio', value='1r', type=MetaType.Text)], bulk=False)
        p1_2.metadatas.set([MetaData(name='folio', value='1v', type=MetaType.Text)], bulk=False)
        p1_3.metadatas.set([MetaData(name='folio', value='2r', type=MetaType.Text)], bulk=False)
        p1_1.add_parent(vol1)
        p1_2.add_parent(vol1)
        p1_3.add_parent(vol1)

        p2_1 = Element.objects.create(corpus=corpus, type=ElementType.Page, name="Volume 2, page 1r", zone=z4)
        p2_2 = Element.objects.create(corpus=corpus, type=ElementType.Page, name="Volume 2, page 1v", zone=z5)
        p2_3 = Element.objects.create(corpus=corpus, type=ElementType.Page, name="Volume 2, page 2r", zone=z6)
        p2_1.metadatas.set([MetaData(name='folio', value='1r', type=MetaType.Text)], bulk=False)
        p2_2.metadatas.set([MetaData(name='folio', value='1v', type=MetaType.Text)], bulk=False)
        p2_3.metadatas.set([MetaData(name='folio', value='2r', type=MetaType.Text)], bulk=False)
        p2_1.add_parent(vol2)
        p2_2.add_parent(vol2)
        p2_3.add_parent(vol2)

        # Create transcriptions on images of volume 1
        for page in (p1_1, p1_2, p1_3):
            for word, pos in [("PARIS", 100), ("ROY", 400), ("DATUM", 700)]:
                Transcription.objects.create(
                    element=page,
                    source=source,
                    text=word,
                    type=TranscriptionType.Word,
                    zone=makezone(page.zone.image, pos, pos + 100),
                    score=1.0,
                )

        # Create 5 acts on volume 1
        act1 = Element.objects.create(corpus=corpus, type=ElementType.Act, name="Act 1")
        act2 = Element.objects.create(corpus=corpus, type=ElementType.Act, name="Act 2")
        act3 = Element.objects.create(corpus=corpus, type=ElementType.Act, name="Act 3")
        act4 = Element.objects.create(corpus=corpus, type=ElementType.Act, name="Act 4")
        act5 = Element.objects.create(corpus=corpus, type=ElementType.Act, name="Act 5")
        act1.metadatas.set([MetaData(name='number', value='1', type=MetaType.Text)], bulk=False)
        act2.metadatas.set([MetaData(name='number', value='2', type=MetaType.Text)], bulk=False)
        act3.metadatas.set([MetaData(name='number', value='3', type=MetaType.Text)], bulk=False)
        act4.metadatas.set([MetaData(name='number', value='4', type=MetaType.Text)], bulk=False)
        act5.metadatas.set([MetaData(name='number', value='5', type=MetaType.Text)], bulk=False)
        act1.add_parent(vol1)
        act2.add_parent(vol1)
        act3.add_parent(vol1)
        act4.add_parent(vol1)
        act5.add_parent(vol1)

        # Create surfaces on the 3 pages of volume 1
        # Page 1 [A B] Page 2 [C] Page 3 [D E F]
        sa = Element.objects.create(corpus=corpus, type=ElementType.Surface,
                                    name="Surface A", zone=makezone(img1, 0, 600))
        sb = Element.objects.create(corpus=corpus, type=ElementType.Surface,
                                    name="Surface B", zone=makezone(img1, 600, 1000))
        sc = Element.objects.create(corpus=corpus, type=ElementType.Surface,
                                    name="Surface C", zone=makezone(img2, 0, 1000))
        sd = Element.objects.create(corpus=corpus, type=ElementType.Surface,
                                    name="Surface D", zone=makezone(img3, 0, 300))
        se = Element.objects.create(corpus=corpus, type=ElementType.Surface,
                                    name="Surface E", zone=makezone(img3, 300, 600))
        sf = Element.objects.create(corpus=corpus, type=ElementType.Surface,
                                    name="Surface F", zone=makezone(img3, 600, 1000))
        # Act 1 [A] Act 2 [B C] Act 3 [D] Act 4 [E] Act 5 [F]
        sa.add_parent(act1)
        sb.add_parent(act2)
        sc.add_parent(act2)
        sd.add_parent(act3)
        se.add_parent(act4)
        sf.add_parent(act5)

        # Create OAuth credentials for a user
        creds = user.credentials.create(
            provider_name='gitlab',
            provider_url='https://somewhere',
            token='oauth-token',
        )

        # Create a repository
        repo = creds.repos.create(
            url='http://gitlab/repo',
            hook_token='hook-token',
            corpus=corpus,
            provider_name='GitLabProvider',
        )

        # Create a revision on the repository
        repo.revisions.create(
            hash='42',
            ref='refs/heads/master',
            message='a',
            author='me',
        )
