from django.conf import settings
from django.db.models.signals import pre_delete
from ponos.management.base import PonosCommand
from arkindex.project.argparse import CorpusArgument
from arkindex.project.tools import disconnect_signal
from arkindex.dataimport.models import Revision, Event
from arkindex.documents.models import Element, ElementPath, Transcription, Classification, MetaData, EntityLink
from arkindex.documents.signals import pre_delete_handler
from arkindex.images.models import Zone
import logging

logging.basicConfig(
    format='%(asctime)s [%(levelname)s] %(message)s',
    level=logging.INFO,
)

logger = logging.getLogger(__name__)


class Command(PonosCommand):
    help = 'Delete a large corpus'
    docker_image = settings.ARKINDEX_APP_IMAGE
    base_recipe = settings.PONOS_RECIPE

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            'corpus',
            type=CorpusArgument(),
            help='ID or part of the name of a corpus to delete.',
        )
        parser.add_argument(
            '--dry-run',
            action='store_true',
            default=False,
            help='Print every action that would be performed, without actually deleting anything',
        )
        parser.add_argument(
            '--batch-size',
            type=int,
            default=10000,
            help='Size of each batch of elements to delete',
        )

    def run(self, corpus=None, dry_run=False, batch_size=1000, **kwargs):
        logger.info('Running corpus deletion{} for {}'.format(' (dry run)' if dry_run else '', corpus.name))
        logger.info('Fetching element IDs')
        element_ids = list(corpus.elements.values_list('id', flat=True))
        element_count = len(element_ids)
        logger.info('{} elements in corpus'.format(element_count))

        # Delete related models manually to minimize execution time and memory usage
        querysets = [
            Event.objects.filter(element_id__in=element_ids),
            corpus.imports.all(),
            Revision.objects.filter(repo__corpus_id=corpus.id),
            corpus.repos.all(),
            corpus.files.all(),
            EntityLink.objects.filter(role__corpus_id=corpus.id),
            corpus.entities.all(),
            corpus.roles.all(),
            MetaData.objects.filter(element_id__in=element_ids),
            Classification.objects.filter(element_id__in=element_ids),
            Transcription.objects.filter(element_id__in=element_ids),
            Zone.objects.filter(elements__in=element_ids),
            ElementPath.objects.filter(element_id__in=element_ids),
        ]

        for queryset in querysets:
            logger.info('{} {} {}'.format(
                'Would delete' if dry_run else 'Deleting',
                queryset.count(),
                queryset.model.__name__,
            ))
            if not dry_run:
                queryset.delete()

        if dry_run:
            logger.info('Would delete {} elements'.format(element_count))
            logger.info('Would delete corpus {}'.format(corpus.name))
            return

        # Temporarily disconnect the pre_delete signal for Elements that removes paths
        # Makes the deletion process 6 times faster
        with disconnect_signal(pre_delete, sender=Element, receiver=pre_delete_handler):
            logger.info('Deleting {} elements'.format(element_count))
            deleted = 0
            for i in range(0, element_count, batch_size):
                elts = Element.objects.filter(id__in=element_ids[i:i+batch_size])
                deleted += elts.count()
                elts.delete()
                logger.info('Deleted {} elements out of {} ({: >3}%)'.format(
                    deleted,
                    element_count,
                    int(100.0 * deleted / element_count),
                ))

        logger.info('Deleting corpus {}'.format(corpus.name))
        corpus.delete()
        logger.info('Successfully removed corpus {}'.format(corpus.name))
