from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import MetaType
from arkindex.documents.models import Element, ElementType, Page, Act, Corpus, MetaData, Region, DataSource
from arkindex.images.serializers import ZoneSerializer, ImageSerializer
from arkindex.images.models import Image
from arkindex.documents.serializers.light import \
    CorpusLightSerializer, ElementLightSerializer, EntityLightSerializer, RegionLightSerializer, MetaDataLightSerializer
from arkindex.documents.serializers.ml import ClassificationSerializer, TranscriptionSerializer
from arkindex.dataimport.models import EventType
from arkindex.project.serializer_fields import EnumField, PolygonField, MLToolField
from arkindex.project.polygon import Polygon
from arkindex.project.aws import S3FileStatus
from collections import defaultdict


class MetaDataSerializer(MetaDataLightSerializer):
    """
    Serialises some Metadata for any Element
    """
    entity = EntityLightSerializer()

    class Meta:
        model = MetaData
        fields = MetaDataLightSerializer.Meta.fields + ('entity', )


class CorpusSerializer(serializers.ModelSerializer):
    """
    Serialize a corpus
    """
    rights = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'description',
            'public',
            'rights',
        )

    def get_rights(self, corpus):
        return map(lambda x: x.value, corpus.get_acl_rights(self.context['request'].user))

    def create(self, validated_data):
        c = Corpus.objects.create(**validated_data)
        c.corpus_right.create(
            user=self.context['request'].user,
            can_write=True,
            can_admin=True,
        )
        return c


class ElementSlimSerializer(serializers.ModelSerializer):
    """
    Fully Serialises a document
    """
    type = EnumField(ElementType, read_only=True)
    corpus = CorpusLightSerializer(read_only=True)
    thumbnail_url = serializers.URLField(source='thumbnail.s3_url', read_only=True)
    thumbnail_put_url = serializers.URLField(source='thumbnail.s3_put_url', read_only=True)

    def get_s3_put_url(self, element):
        user = self.context['request'].user
        if element.zone.image.server.is_local and user.is_authenticated and (user.is_admin or user.is_internal):
            return element.thumbnail.s3_put_url

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
            'corpus',
            'thumbnail_url',
            'thumbnail_put_url',
        )


class ElementSerializer(ElementSlimSerializer):
    """
    Serialize an element with its metadata and classifications
    """

    metadata = serializers.SerializerMethodField()
    classifications = ClassificationSerializer(many=True)

    class Meta:
        model = Element
        fields = ElementSlimSerializer.Meta.fields + ('metadata', 'classifications')

    def create(self, *args, **kwargs):
        instance = super().create(*args, **kwargs)
        instance.events.create(type=EventType.Addition)
        return instance

    def update(self, *args, **kwargs):
        instance = super().update(*args, **kwargs)
        instance.events.create(type=EventType.Edit)
        return instance

    def get_metadata(self, obj):
        return MetaDataLightSerializer(
            list(obj.metadatas.all().filter(entity__isnull=True).exclude(type=MetaType.Entity)),
            many=True
        ).data


class ElementNeighborsSerializer(serializers.Serializer):
    position = serializers.IntegerField()
    element = ElementLightSerializer()
    parents = serializers.ListField(
        child=ElementLightSerializer()
    )


class ActSerializer(serializers.ModelSerializer):
    """
    Serialize an act with its parents and children
    and metadatas
    """

    parents = serializers.ListField(
        child=serializers.ListField(
            child=ElementSlimSerializer()
        ),
        source='parent_elements',
        read_only=True,
    )
    children = ElementSlimSerializer(source='child_elements', read_only=True, many=True)
    metadata = serializers.SerializerMethodField()
    classifications = ClassificationSerializer(many=True)

    def get_metadata(self, obj):
        return MetaDataLightSerializer(
            obj.metadatas.all().filter(entity__isnull=True).exclude(type=MetaType.Entity),
            many=True
        ).data

    class Meta:
        model = Act
        fields = (
            'id',
            'name',
            'parents',
            'children',
            'metadata',
            'classifications'
        )
        read_only_fields = ('id', )


class PageSerializer(serializers.ModelSerializer):
    """
    Serialises a Page
    """
    image = ImageSerializer(source='zone.image')
    corpus = CorpusLightSerializer()

    parents = serializers.ListField(
        child=serializers.ListField(
            child=ElementSlimSerializer(),
        ),
        source='parent_elements',
        read_only=True,
    )

    transcriptions = TranscriptionSerializer(
        source='page_transcriptions',
        many=True,
    )

    class Meta:
        model = Page
        fields = (
            'id',
            'name',
            'image',
            'corpus',
            'parents',
            'transcriptions',
        )


class SurfaceSerializer(serializers.ModelSerializer):
    """
    Serialize a surface with its page, image and zone
    """

    zone = ZoneSerializer()
    page = PageSerializer()

    class Meta:
        model = Element
        fields = (
            'id',
            'name',
            'page',
            'zone',
        )


class ElementCreateSerializer(ElementLightSerializer):
    """
    Serialize an Element with a possible parent and image
    """

    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())
    image = serializers.PrimaryKeyRelatedField(
        queryset=Image.objects.all().prefetch_related('server'),
        required=False
    )
    parent = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none(), required=False)
    # TODO: Replace with a list of MetaDataLightSerializer
    metadata = serializers.DictField(child=serializers.CharField(), required=False)

    class Meta(ElementLightSerializer.Meta):
        model = Element
        # handle specific cases ()
        fields = ElementLightSerializer.Meta.fields + ('image', 'corpus', 'parent', 'metadata')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora
        self.fields['parent'].queryset = Element.objects.filter(corpus__in=corpora)

    def validate(self, data):
        errors = defaultdict(list)

        data = super().validate(data)
        parent = data.get('parent')
        if parent and parent.corpus != data['corpus']:
            errors['corpus'].append(
                'Parent and child must be in the same corpus'
            )
        image = data.get('image')
        if image and image.status != S3FileStatus.Checked:
            errors['image'].append(
                'Image is not checked. Try to upload a valid IIIF image.'
            )

        # Limit Elements creation cases
        if parent and parent.type != ElementType.Volume:
            errors['parent'].append(
                'Parent type should be a Volume'
            )
        elif parent and data['type'] not in (ElementType.Page, ElementType.Act):
            errors['type'].append(
                'Only a Page or an Act type child can be created'
            )
        if not parent and data['type'] != ElementType.Volume:
            errors['parent'].append(
                'Only a Volume can be created without any parent'
            )

        if errors:
            raise ValidationError(errors)
        return data

    def create(self, validated_data):
        image = validated_data.pop('image', None)
        parent = validated_data.pop('parent', None)
        metadata = validated_data.pop('metadata', {})

        element = Element(**validated_data)
        element.full_clean()

        if image:
            # Create a new zone containing the image
            zone = image.zones.create(polygon=Polygon.from_coords(0, 0, image.width, image.height))
            # update Element zone
            element.zone = zone

        element.save()
        # Make relation to parent
        if parent:
            element.add_parent(parent)

        for name, value in metadata.items():
            element.metadatas.create(name=name, type=MetaType.Text, value=value)

        return element


class RegionSerializer(RegionLightSerializer):
    elements = ElementLightSerializer(many=True)

    class Meta(RegionLightSerializer.Meta):
        fields = RegionLightSerializer.Meta.fields + ('elements', )


class RegionCreateSerializer(serializers.Serializer):
    id = serializers.UUIDField(read_only=True)
    class_name = serializers.CharField(source='ml_class')
    confidence = serializers.FloatField(min_value=0, max_value=1)
    polygon = PolygonField()
    analyser = MLToolField(MLToolType.DLAnalyser)

    def validate_class_name(self, name):
        assert 'element' in self.context, 'Missing associated element'
        return self.context['element'].corpus.ml_classes.get_or_create(name=name)[0]

    def save(self):
        region, _ = Region.objects.get_or_create(
            ml_class=self.validated_data['ml_class'],
            zone=self.context['element'].zone.image.zones.get_or_create(
                polygon=self.validated_data['polygon'],
            )[0],
            confidence=self.validated_data['confidence'],
            source=DataSource.from_ml_tool(self.validated_data['analyser']),
        )
        region.elements.add(self.context['element'])
        self.validated_data['id'] = region.id
        return region
