from rest_framework import serializers
from arkindex.documents.models import Element, ElementType, Corpus, Page, Entity, Region, MetaData
from arkindex_common.enums import EntityType, MetaType
from arkindex.documents.dates import DateType
from arkindex.documents.serializers.ml import DataSourceSerializer, MLClassSerializer
from arkindex.images.serializers import ImageSerializer, ZoneSerializer
from arkindex.dataimport.serializers.git import RevisionSerializer
from arkindex.project.serializer_fields import EnumField


class InterpretedDateSerializer(serializers.Serializer):
    """
    Serialize a list of interpreted dates linked to a metadata
    """
    type = EnumField(DateType)
    year = serializers.IntegerField(min_value=0)
    month = serializers.IntegerField(min_value=1, max_value=12, allow_null=True)
    day = serializers.IntegerField(min_value=1, max_value=31, allow_null=True)

    class Meta:
        fields = (
            'year',
            'month',
            'day',
            'type',
        )


class ElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Element
    """
    type = EnumField(ElementType)

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )


class PageLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Page
    """
    image = ImageSerializer(source='zone.image')
    type = EnumField(ElementType)

    class Meta:
        model = Page
        fields = (
            'id',
            'type',
            'name',
            'image',
        )


class CorpusLightSerializer(serializers.ModelSerializer):
    """
    Serializes a Corpus
    """

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'public'
        )


class EntityLightSerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EnumField(EntityType)
    dates = InterpretedDateSerializer(many=True, source='get_dates', required=False, allow_null=True)
    metas = serializers.HStoreField(child=serializers.CharField(), required=False, allow_null=True)
    source = DataSourceSerializer()

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'dates',
            'source'
        )


class RegionLightSerializer(serializers.ModelSerializer):
    """
    Serialize a region without its related elements
    """
    ml_class = MLClassSerializer()
    source = DataSourceSerializer()
    zone = ZoneSerializer()

    class Meta:
        model = Region
        fields = (
            'id',
            'source',
            'ml_class',
            'confidence',
            'zone',
        )


class MetaDataLightSerializer(serializers.ModelSerializer):
    """
    Serialize a metadata without its entity
    """
    type = EnumField(MetaType)
    revision = RevisionSerializer()
    dates = InterpretedDateSerializer(many=True, source='get_dates')

    class Meta:
        model = MetaData
        fields = (
            'id',
            'type',
            'name',
            'value',
            'revision',
            'dates',
        )
