from unittest.mock import patch
from pathlib import Path
from arkindex_common.enums import MetaType
from arkindex.project.tests import FixtureTestCase
from arkindex.project.polygon import Polygon
from arkindex.documents.models import Element, ElementType, Act
from arkindex.documents.acts import ActsImporter

ACT_SAMPLES = Path(__file__).resolve().parent / 'act_samples'


class TestActsImporter(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.img4 = cls.imgsrv.images.get(path='img4')
        cls.img5 = cls.imgsrv.images.get(path='img5')
        cls.img6 = cls.imgsrv.images.get(path='img6')
        cls.volume = cls.corpus.elements.get(name='Volume 2')

    def test_import(self):
        self.assertFalse(
            Element.objects
                   .get_descending(self.volume.id)
                   .filter(type__in=(ElementType.Act, ElementType.Surface))
                   .exists()
        )
        importer = ActsImporter(ACT_SAMPLES / 'volume 2.csv', self.corpus)
        importer.run()
        self.assertEqual(importer.created_acts_count, 3)
        self.assertEqual(importer.created_surfaces_count, 3)
        self.assertEqual(Act.objects.get_descending(self.volume.id).count(), 3)
        act_1, act_2bis, act_3 = Act.objects.get_descending(self.volume.id).order_by('name')

        self.assertEqual(act_1.name, 'Act 1')
        metadata = act_1.metadatas.get()
        self.assertEqual(metadata.name, 'number')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertEqual(metadata.value, '1')
        self.assertFalse(Element.objects.get_descending(act_1.id).filter(type=ElementType.Surface).exists())

        self.assertEqual(act_2bis.name, 'Act 2bis')
        metadata = act_2bis.metadatas.get()
        self.assertEqual(metadata.name, 'number')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertEqual(metadata.value, '2bis')
        self.assertEqual(Element.objects.get_descending(act_2bis.id).filter(type=ElementType.Surface).count(), 2)
        surf1, surf2 = Element.objects.get_descending(act_2bis.id).filter(type=ElementType.Surface)

        self.assertEqual(surf1.name, 'Surface 1 on act 2bis')
        self.assertEqual(surf1.zone.image, self.img4)
        self.assertEqual(surf1.zone.polygon, Polygon.from_coords(0, 0, 1000, 1000))

        self.assertEqual(surf2.name, 'Surface 2 on act 2bis')
        self.assertEqual(surf2.zone.image, self.img5)
        self.assertEqual(surf2.zone.polygon, Polygon.from_coords(0, 0, 1000, 500))  # Half of the image

        self.assertEqual(act_3.name, 'Act 3')
        metadata = act_3.metadatas.get()
        self.assertEqual(metadata.name, 'number')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertEqual(metadata.value, '3')
        surface = Element.objects.get_descending(act_3.id).get(type=ElementType.Surface)
        self.assertEqual(surface.zone.image, self.img5)
        self.assertEqual(surface.zone.polygon, Polygon.from_coords(0, 500, 1000, 500))  # Second half

    def test_epic_fail(self):
        importer = ActsImporter(ACT_SAMPLES / 'volume 2.csv', self.corpus)
        with patch.object(importer, 'handle_line') as mock:
            # Fake an error at every line of the CSV
            mock.side_effect = ValueError
            with self.assertRaisesRegex(AssertionError, 'No acts were imported'):
                importer.run()
            self.assertEqual(mock.call_count, 6)
