from django.urls import reverse
from rest_framework import status
from arkindex_common.enums import MetaType, TranscriptionType, EntityType
from arkindex.documents.models import Element, ElementType, DataSource, \
    Page, Act, Corpus, Entity, MLClass, Classification
from arkindex.images.models import ImageServer
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.aws import S3FileStatus
from arkindex.users.models import User


class TestElementsAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.vol = Element.objects.get(name='Volume 1', type=ElementType.Volume, corpus=cls.corpus)
        cls.image = ImageServer.objects.local.images.create(
            path="kingdom/far/away",
            status=S3FileStatus.Checked
        )

    def setUp(self):
        self.page = self.corpus.elements.get(type=ElementType.Page, name='Volume 1, page 1r')
        self.folio_metadata = self.page.metadatas.get()
        self.metadata = self.page.metadatas.create(
            name='meta',
            type=MetaType.Text,
            value='text_metadata'
        )

    def test_get_element(self):
        data_source = DataSource.objects.get(slug='test')
        ml_class = MLClass.objects.create(name='text', corpus=self.corpus)
        classification = Classification.objects.create(element=self.vol, source=data_source, ml_class=ml_class)

        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.vol.id))
        self.assertEqual(data['name'], 'Volume 1')
        self.assertEqual(data['corpus']['id'], str(self.corpus.id))
        self.assertEqual(data['type'], 'volume')

        self.assertEqual(len(data['classifications']), 1)
        cl = data['classifications'][0]
        self.assertEqual(cl['id'], str(classification.id))
        self.assertEqual(cl['source']['id'], str(classification.source.id))
        self.assertEqual(cl['ml_class']['id'], str(classification.ml_class.id))

    def test_patch_element_unverified(self):
        """
        Check patching an element with an unverified email is not possible
        """
        user = User.objects.get_by_natural_key('demo@demo.fr')
        self.assertFalse(user.verified_email)
        self.client.force_login(user)

        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_patch_element_acl(self):
        """
        Check patching an element as anonymous user is not possible
        """
        self.user.corpus_right.get(corpus=self.corpus).delete()
        self.assertTrue(self.user.verified_email)
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_element(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['name'], 'Untitled (2)')
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.name, 'Untitled (2)')

    def make_create_request(self, name='default', corpus=None, elt_type='volume', **options):
        request = {
            'path': reverse('api:elements-create'),
            'data': {
                'type': elt_type,
                'name': name,
                'corpus': corpus or str(self.corpus.id),
            },
            'format': 'json',
        }
        # Optional parameters
        request['data'].update(options)
        return request

    def test_create_element_not_verified(self):
        self.assertEqual(
            self.client.post(**self.make_create_request('no')).status_code,
            status.HTTP_403_FORBIDDEN
        )

    def test_create_element_volume(self):
        # Create a Volume
        self.client.force_login(self.user)
        request = self.make_create_request('my new volume')
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        volume = Element.objects.get(id=data['id'])
        self.assertDictEqual(
                data,
                {
                    'id': str(volume.id),
                    'corpus': str(volume.corpus.id),
                    'name': volume.name,
                    'type': volume.type.value,
                }
        )
        self.assertEqual(volume.name, 'my new volume')

    def test_create_volume_page(self):
        # Add a page with an Image to an existing volume
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams',
            image=str(self.image.id),
            metadata={'folio': 'castle 001'}
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = Page.objects.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'corpus': str(page.corpus.id),
                'name': page.name,
                'type': page.type.value,
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams')
        self.assertEqual(page.metadatas.count(), 1)
        metadata = page.metadatas.get()
        self.assertEqual(metadata.name, 'folio')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertEqual(metadata.value, 'castle 001')

    def test_create_volume_act(self):
        # Create an volume child of type Act
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            name='Castle story',
            elt_type='act',
            metadata={
                'folio': 'castle_story',
                'number': '42',
            }
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Act.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.metadatas.count(), 2)
        folio_metadata, number_metadata = act.metadatas.order_by('name')
        self.assertEqual(number_metadata.name, 'number')
        self.assertEqual(number_metadata.type, MetaType.Text)
        self.assertEqual(number_metadata.value, '42')
        self.assertEqual(folio_metadata.name, 'folio')
        self.assertEqual(folio_metadata.type, MetaType.Text)
        self.assertEqual(folio_metadata.value, 'castle_story')

    def test_create_parent_different_corpus(self):
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        request = self.make_create_request(
            parent=str(self.vol.id),
            corpus=str(new_corpus.id),
            elt_type='act',
            metadata={
                'folio': 'new',
                'number': '43',
            },
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn('corpus', response.json())

    def test_create_wrong_image(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id),
            metadata={'folio': 'new'},
        )
        self.image.status = S3FileStatus.Error
        self.image.save()
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn('image', response.json())

    def test_create_wrong_schema(self):
        # Parent is not a volume
        self.client.force_login(self.user)
        request = self.make_create_request(
                parent=str(self.page.id),
                elt_type='act',
                metadata={'folio': 'new'},
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn('parent', response.json())

    def test_get_transcriptions(self):
        self.page.transcriptions.all().delete()
        transcriptions = []
        for i in range(10):
            transcriptions.append(self.page.transcriptions.create(
                source_id=DataSource.objects.get(slug='test').id,
                type=TranscriptionType.Word
            ))
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 10)
        self.assertEqual(len(transcriptions), 10)
        # Wrong transcription type
        url = reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)})
        url += '?type=potato'
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_elements_filter_name(self):
        nameSelected = 'ol'
        response = self.client.get(reverse('api:elements'), data={'name': nameSelected})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertIn(nameSelected, result['name'])

    def test_list_elements_filter_type(self):
        response = self.client.get(reverse('api:elements'), data={'type': 'act'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertEqual(result['type'], 'act')

    def test_list_elements_filter_wrong_type(self):
        response = self.client.get(reverse('api:elements'), data={'type': 'document'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_element_metadata(self):
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['metadata']
        self.assertListEqual(results, [
            {
                'dates': [],
                'id': str(self.folio_metadata.id),
                'name': 'folio',
                'revision': None,
                'type': 'text',
                'value': '1r',
            },
            {
                'id': str(self.metadata.id),
                'name': str(self.metadata.name),
                'type': self.metadata.type.value,
                'value': self.metadata.value,
                'dates': [],
                'revision': None
            },
        ])

    def test_list_element_metadata_with_entity(self):
        # Link an entity to metadata, db requests number should remains same
        entity = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name='Marc',
            source_id=DataSource.objects.get(slug='test').id
        )
        self.metadata.entity = entity
        self.metadata.save()
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['metadata']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'folio')

    def test_list_element_metadata_wrong_acl(self):
        self.page.corpus = self.private_corpus
        self.page.save()
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def build_tree(self, tree, element_type=ElementType.Volume):
        """
        Build Element and ElementPath instances from a tree described by
        a dict associating elements with their direct ancestors.

        {'child': ['mother', 'father'], 'grandchild': 'child'}

        Returns a dict associating element names with created Elements.
        """
        assert isinstance(tree, dict)

        def parse_value(val):
            if isinstance(val, str):
                return [val]
            return val

        # List all elements
        # Pick all the keys, then add elements in the tree's values
        # Removes the need to add 'element': [] for root elements
        names = set(tree.keys()).union(*[parse_value(val) for val in tree.values()])
        # Create elements
        elements = {
            name: Element.objects.create(
                corpus=self.corpus,
                name=name,
                type=element_type,
            )
            for name in names
        }
        # Create paths
        for child_name, value in tree.items():
            parent_names = parse_value(value)
            for parent_name in parent_names:
                elements[child_name].add_parent(elements[parent_name])
        return elements

    def test_element_neighbors(self):
        elements = self.build_tree({
            'Y': 'Z',
            'B': 'X',
            'A': ['X', 'Y'],
            'C': 'X',
        })
        # Neighbors of A with a different type
        element = Element.objects.create(corpus=self.corpus, type=ElementType.Page, name='D')
        element.add_parent(elements['Y'])
        element = Element.objects.create(corpus=self.corpus, type=ElementType.Page, name='E')
        element.add_parent(elements['Y'])

        # Neighbors of A with the same type
        element = Element.objects.create(corpus=self.corpus, type=ElementType.Volume, name='F')
        element.add_parent(elements['Y'])

        response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['A'].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertListEqual(
            results,
            [{
                'position': 0,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': elements['B'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 2,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['Y'].id),
                    'type': 'volume',
                    'name': elements['Y'].name
                }, {
                    'id': str(elements['Z'].id),
                    'type': 'volume',
                    'name': elements['Z'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(element.id),
                    'type': 'volume',
                    'name': element.name
                },
                'parents': [{
                    'id': str(elements['Y'].id),
                    'type': 'volume',
                    'name': elements['Y'].name
                }, {
                    'id': str(elements['Z'].id),
                    'type': 'volume',
                    'name': elements['Z'].name
                }]
            }]
        )

    def test_n_element_neighbors(self):
        elements = self.build_tree({
            'A': 'W',
            'B': 'W',
            'C': 'W',
        })

        response = self.client.get(
            reverse('api:elements-neighbors', kwargs={'pk': str(elements['A'].id)}),
            data={'n': 2})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertListEqual(
            results,
            [{
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': 'B'
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 2,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }]
        )
