from unittest.mock import MagicMock
from django.contrib.auth.models import AnonymousUser
from django.urls import reverse
from rest_framework import status
from elasticsearch_dsl.connections import connections
from arkindex_common.enums import EntityType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.elastic import ESTranscription
from arkindex.documents.models import Transcription, Element, Corpus, DataSource, MLToolType
import uuid


class TestSearchAPI(FixtureAPITestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.previous_connection = connections.get_connection('default')
        cls.es_mock = MagicMock()
        connections.add_connection('default', cls.es_mock)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        connections.add_connection('default', cls.previous_connection)

    def setUp(self):
        self.es_mock.reset_mock()

    def build_es_response(self, hits):
        return {
            "hits": {
                "total": len(hits),
                "max_score": None,
                "hits": hits
            },
            "_shards": {
                "total": 5,
                "failed": 0,
                "skipped": 0,
                "successful": 5
            },
            "took": 42,
            "timed_out": False
        }

    def make_transcription_hit(self, ts):
        return {
            "_id": str(ts.id.hex),
            "_score": None,
            "sort": [
                ts.score
            ],
            "_index": "transcriptions",
            "_type": "transcription",
            "_source": ESTranscription.from_model(ts).to_dict()
        }

    def make_nested_transcription_hit(self, ts):
        return {
            "_source": {
                "text": ts.text,
                "id": str(ts.id),
                "score": ts.score,
            },
            "_score": ts.score,
            "_nested": {
                "field": "transcriptions",
                "offset": 1337
            }
        }

    def make_element_hit(self, elt, ts, score=1.0):
        return {
            "_score": score,
            "_type": 'element',
            "_id": str(elt.id.hex),
            # TODO test date ranges in a query
            '_source': {'date_range': []},
            "_index": 'elements',
            "inner_hits": {
                "transcriptions": {
                    "hits": {
                        "total": len(ts),
                        "hits": list(map(self.make_nested_transcription_hit, ts)),
                        "max_score": 1337,
                    }
                }
            }
        }

    def make_entity_hit(self, entity):
        return {
            "_id": str(entity.id),
            "_index": "entities",
            "_type": "entity",
        }

    def test_element_search(self):
        elt = Element.objects.get(name="Volume 1, page 1r")
        ts = Transcription.objects.filter(text="PARIS", zone__image__path='img1')

        self.es_mock.count.return_value = {'count': 1}
        self.es_mock.search.return_value = self.build_es_response(
            [self.make_element_hit(elt, ts), ],
        )

        response = self.client.get(reverse('api:element-search'), {'q': "paris"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()["results"]
        self.assertEqual(len(results), 1)
        result = results[0]

        self.assertEqual(result['id'], str(elt.id))
        self.assertCountEqual(
            [t['id'] for t in result['transcriptions']],
            map(str, ts.values_list('id', flat=True)),
        )
        self.assertEqual(result['total_transcriptions'], len(ts))

        args, kwargs = self.es_mock.search.call_args
        self.assertTupleEqual(args, ())
        self.assertCountEqual(kwargs.keys(), ['body', 'index', 'doc_type'])
        self.assertListEqual(kwargs['index'], ['elements'])
        self.assertListEqual(kwargs['doc_type'], ['element'])

        self.assertCountEqual(kwargs['body'].keys(), ['_source', 'from', 'size', 'query', 'sort'])
        self.assertListEqual(kwargs['body']['_source'], ['date_range'])
        self.assertEqual(kwargs['body']['from'], 0)
        self.assertEqual(kwargs['body']['size'], 1)

        self.assertCountEqual(
            kwargs['body']['query']['bool']['filter'][0]['terms']['corpus'],
            map(str, Corpus.objects.readable(AnonymousUser()).values_list('id', flat=True)),
        )

        nested = kwargs['body']['query']['bool']['must'][0]['nested']
        self.assertEqual(nested['score_mode'], 'sum')
        self.assertEqual(nested['path'], 'transcriptions')
        self.assertIn('inner_hits', nested)

        function_score = nested['query']['function_score']
        self.assertListEqual(function_score['functions'], [
            {
                "field_value_factor": {
                    "field": "transcriptions.score",
                }
            }
        ])
        self.assertIsInstance(function_score['query']['bool']['must'], list)

        self.assertTrue(all(len(cond.keys()) == 1 for cond in function_score['query']['bool']['must']))
        conditions = function_score['query']['bool']['must']
        self.assertCountEqual(conditions, [
            {'simple_query_string': {
                'query': 'paris',
                'fields': ['transcriptions.text'],
            }},
            {'range': {
                'transcriptions.score': {
                    'gte': 0.0,
                }
            }},
        ])

    def test_iiif_transcription_search(self):
        # Filter to only get transcriptions from volume 1
        unfiltered = Transcription.objects.filter(text="PARIS")
        expected = Transcription.objects.filter(text="PARIS", zone__image__path__in=['img1', 'img2', 'img3'])
        vol = Element.objects.get(name='Volume 1')

        self.es_mock.count.return_value = {'count': len(unfiltered)}
        self.es_mock.search.return_value = self.build_es_response(
            list(map(self.make_transcription_hit, unfiltered))
        )

        response = self.client.get(reverse('api:ts-search-manifest', kwargs={'pk': str(vol.id)}), {'q': 'paris'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(data['@context'], "http://iiif.io/api/search/0/context.json")
        self.assertEqual(data['@type'], 'sc:AnnotationList')
        self.assertEqual(data['startIndex'], 0)
        self.assertEqual(data['within']['@type'], 'sc:Layer')
        self.assertEqual(data['within']['total'], len(expected))

        hits = data['hits']
        self.assertTrue(all(hit['@type'] == 'search:Hit' for hit in hits))
        self.assertTrue(all(hit['match'] == 'PARIS' for hit in hits))

        resources = data['resources']
        self.assertTrue(all(res['@type'] == 'oa:Annotation' for res in resources))
        self.assertTrue(all(res['motivation'] == 'sc:painting' for res in resources))
        self.assertTrue(all(res['resource']['@type'] == 'cnt:ContentAsText' for res in resources))
        self.assertTrue(all(res['resource']['format'] == 'text/plain' for res in resources))
        self.assertTrue(all(res['resource']['chars'] == 'PARIS' for res in resources))

        args, kwargs = self.es_mock.search.call_args
        self.assertTupleEqual(args, ())
        self.assertCountEqual(kwargs.keys(), ['body', 'index', 'doc_type'])
        self.assertListEqual(kwargs['index'], ['transcriptions'])
        self.assertListEqual(kwargs['doc_type'], ['transcription'])

        self.assertCountEqual(kwargs['body'].keys(), ['query', 'from', 'size'])
        self.assertEqual(kwargs['body']['from'], 0)
        self.assertEqual(kwargs['body']['size'], 10000)
        self.assertIsInstance(kwargs['body']['query']['bool']['must'], list)
        self.assertTrue(all(len(cond.keys()) == 1 for cond in kwargs['body']['query']['bool']['must']))
        conditions = kwargs['body']['query']['bool']['must']

        self.assertListEqual(conditions, [{'match': {'text': 'paris'}}])

    def test_entity_search(self):
        source = DataSource.objects.create(type=MLToolType.NER, slug='entity source', internal=True)
        entity_1 = self.corpus.entities.create(type=EntityType.Person, name="an entity", source=source)
        entity_2 = self.corpus.entities.create(type=EntityType.Location, name="somewhere", source=source)
        self.es_mock.count.return_value = {'count': 2}
        self.es_mock.search.return_value = self.build_es_response([
            # Test the ES ordering is preserved by returning entities in non-alphabetical order
            self.make_entity_hit(entity_2),
            self.make_entity_hit(entity_1),
        ])

        response = self.client.get(reverse('api:entity-search'), {'q': 'some query'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()["results"]
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]['id'], str(entity_2.id))
        self.assertEqual(results[1]['id'], str(entity_1.id))

        args, kwargs = self.es_mock.search.call_args
        self.assertTupleEqual(args, ())
        self.assertCountEqual(kwargs.keys(), ['body', 'index', 'doc_type'])
        self.assertListEqual(kwargs['index'], ['entities'])
        self.assertListEqual(kwargs['doc_type'], ['entity'])

        self.assertCountEqual(kwargs['body'].keys(), ['from', 'size', 'query'])
        self.assertEqual(kwargs['body']['from'], 0)
        self.assertEqual(kwargs['body']['size'], 2)

        self.assertIsInstance(kwargs['body']['query']['bool']['must'], list)
        self.assertTrue(all(len(cond.keys()) == 1 for cond in kwargs['body']['query']['bool']['must']))
        conditions = kwargs['body']['query']['bool']['must']
        self.assertListEqual(conditions, [
            {'simple_query_string': {
                'query': 'some query',
                'fields': ['name'],
            }},
        ])

    def test_entity_search_does_not_exist(self):
        entity_id = uuid.uuid4()
        self.assertFalse(self.corpus.entities.filter(id=entity_id).exists())

        self.es_mock.count.return_value = {'count': 1}
        self.es_mock.search.return_value = self.build_es_response([
            {
                "_id": str(entity_id),
                "_index": "entities",
                "_type": "entity",
            }
        ])
        response = self.client.get(reverse('api:entity-search'), {'q': 'some query'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()["results"]
        self.assertEqual(len(results), 0)
