from django.urls import reverse
from unittest.mock import patch
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.polygon import Polygon
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import Page, Transcription, DataSource
import uuid


class TestTranscriptionCreate(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = Page.objects.get(corpus=cls.corpus, zone__image__path='img1')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.src = DataSource.objects.get(slug='test')

    def test_require_login(self):
        response = self.client.post(reverse('api:transcription-create'), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.documents.api.ml.Indexer')
    def test_create_transcription(self, indexer):
        """
        Checks the view creates transcriptions, zones, links, paths and runs ES indexing
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:transcription-create'), format='json', data={
            "type": "word",
            "element": str(self.page.id),
            "source": self.src.id,
            "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
            "text": "NEKUDOTAYIM",
            "score": 0.83,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_ts = Transcription.objects.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(new_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(new_ts.score, 0.83)
        self.assertEqual(new_ts.source, self.src)
        self.assertTrue(self.page.transcriptions.filter(pk=new_ts.id).exists())
        self.assertTrue(indexer.return_value.run_index.called)

    @patch('arkindex.documents.api.ml.Indexer')
    def test_unique_zone(self, indexer):
        """
        Checks the view reuses zones when available
        """
        self.client.force_login(self.user)
        ts = Transcription.objects.get(zone__image__path='img1', text="PARIS")
        response = self.client.post(reverse('api:transcription-create'), format='json', data={
            "type": "word",
            "element": str(self.page.id),
            "source": self.src.id,
            "polygon": ts.zone.polygon.serialize(),
            "text": "GLOUBIBOULGA",
            "score": 0.8,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Transcription.objects.get(text="GLOUBIBOULGA").zone.id, ts.zone.id)

    @patch('arkindex.documents.api.ml.Indexer')
    def test_update_transcription(self, indexer):
        """
        Checks the view updates transcriptions when they already exist
        """
        self.client.force_login(self.user)
        ts = Transcription.objects.get(zone__image__path='img1', text="PARIS")
        self.assertNotEqual(ts.score, 0.99)
        response = self.client.post(reverse('api:transcription-create'), format='json', data={
            "type": "word",
            "element": str(self.page.id),
            "source": ts.source_id,
            "polygon": ts.zone.polygon.serialize(),
            "text": ts.text,
            "score": 0.99,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Transcription.objects.filter(zone__image__path='img1', text=ts.text).count(), 1)
        ts.refresh_from_db()
        self.assertEqual(ts.score, 0.99)

    def test_invalid_data(self):
        """
        Checks the view validates data properly
        (score between 0 and 1, minimum points in polygon, non-negative coordinates, non-existent element)
        """
        self.client.force_login(self.user)

        post_data = {
            "type": "word",
            "element": str(self.page.id),
            "source": self.src.id,
            "polygon": [(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)],
            "text": "NEKUDOTAYIM",
            "score": 0.83,
        }

        # Negative score
        post_data['score'] = -1
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Score over 100%
        post_data['score'] = 2
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Missing element
        post_data['score'] = 0.83
        post_data['element'] = str(uuid.uuid4())
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Not enough polygon points
        post_data['polygon'] = [(0, 0), (100, 100)]
        response = self.client.post(reverse('api:transcription-create'), format='json', data=post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.images.importer.Indexer')
    def test_create_bulk_transcription(self, indexer, ml_get_mock):
        """
        Checks the view creates transcriptions, zones, links, paths and runs ES indexing
        Using bulk_transcriptions
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data={
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_ts = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(new_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(new_ts.score, 0.83)
        self.assertEqual(new_ts.source, self.src)

        page_ts = self.page.transcriptions.get(type=TranscriptionType.Page)
        self.assertEqual(page_ts.text, 'This is a test')
        self.assertEqual(page_ts.score, None)
        self.assertEqual(page_ts.source, self.src)
        self.assertEqual(page_ts.zone, self.page.zone)

        # Indexer called
        self.assertEqual(indexer.return_value.run_index.call_count, 2)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_bulk_transcription_no_zone(self, ml_get_mock):
        self.assertIsNone(self.vol.zone)

        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data={
            "parent": str(self.vol.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
            ]
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.images.importer.Indexer')
    def test_update_bulk_transcription(self, indexer, ml_get_mock):
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }

        self.client.force_login(self.user)
        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        new_ts = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(new_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(new_ts.score, 0.83)
        self.assertEqual(new_ts.source, self.src)

        page_ts = self.page.transcriptions.get(type=TranscriptionType.Page)
        self.assertEqual(page_ts.text, 'This is a test')
        self.assertEqual(page_ts.score, None)
        self.assertEqual(page_ts.source, self.src)
        self.assertEqual(page_ts.zone, self.page.zone)

        # Indexer called
        self.assertEqual(indexer.return_value.run_index.call_count, 2)

        # Update again
        indexer.reset_mock()
        data['transcriptions'] = [
            {
                "type": "word",
                "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                "text": "NEKUDOTAYIM",
                "score": 0.83,
            },
            {
                "type": "word",
                "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                "text": "Hi",
                "score": 0.99,
            },
        ]

        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(indexer.return_value.run_index.call_count, 2)

        # Previous transcriptions should be replaced by two Word and one Page transcription
        self.assertEqual(self.page.transcriptions.count(), 3)

        # This transcription was inserted with the first call.
        # This ensures calling the update API with an existing transcription
        # keeps the existing transcriptions
        ts_1 = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(ts_1.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(ts_1.score, 0.83)
        self.assertEqual(ts_1.source, self.src)

        ts_2 = self.page.transcriptions.get(text="Hi", type=TranscriptionType.Word)
        self.assertEqual(ts_2.zone.polygon, Polygon.from_coords(0, 0, 200, 200))
        self.assertEqual(ts_2.score, 0.99)
        self.assertEqual(ts_2.source, self.src)

        page_ts = self.page.transcriptions.get(type=TranscriptionType.Page)
        self.assertEqual(page_ts.text, '')
