from django.db.models import Q
from rest_framework.generics import CreateAPIView, RetrieveUpdateAPIView
from rest_framework.serializers import ValidationError
from arkindex.project.aws import S3FileStatus
from arkindex.project.permissions import IsVerified
from arkindex.images.models import Image
from arkindex.images.serializers import ImageSerializer, ImageUploadSerializer


class ImageCreate(CreateAPIView):
    """
    Create an image on the local server, ready for upload to S3
    """

    permission_classes = (IsVerified, )
    serializer_class = ImageUploadSerializer

    def perform_create(self, serializer):
        image_hash = serializer.validated_data.get('hash')
        server = serializer.validated_data.get('server')
        path = serializer.validated_data.get('path')
        existing_image = server.images.filter(
            Q(hash=image_hash) | Q(path=path)
        ).first()

        if existing_image:
            error = {
                'id': str(existing_image.id),
            }
            if existing_image.hash == image_hash:
                error['hash'] = ['Image with this hash already exists']
            if existing_image.path == path:
                error['path'] = ['Image with this path already exists']
            raise ValidationError(error)

        width = serializer.validated_data.get('width')
        height = serializer.validated_data.get('height')
        if width and height and self.request.user.is_internal:
            # Async tasks may send the image's width and height;
            # we will assume that the image was checked by the task
            serializer.validated_data['status'] = S3FileStatus.Checked

        return serializer.save()


class ImageRetrieve(RetrieveUpdateAPIView):
    """
    Get information about an image or trigger a check
    """

    permission_classes = (IsVerified, )
    serializer_class = ImageSerializer
    queryset = Image.objects.all()
